/*
	descrypt.c
	Copyright (C) 2000 David Fifield

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <getopt.h>
#include <sys/stat.h>
#include "descrypt.h"
#include "des.h"
#include "des_mode.h"

int main(int argc, char *argv[])
{
	struct prog_option_t prog_option = {
		NULL,
		DES_ENCRYPT,
		NULL,
		MODE_CBC,
		NULL,
		0
	};
	int (*des_mode[])(FILE *out_file, FILE *in_file,
                      unsigned char key[LONG_KEY_SIZE],
                      unsigned int crypt_dir) = {
		des_mode_ecb,
		des_mode_cbc,
		des_mode_ofb,
		des_mode_cfb
	};
	FILE *in_file, *key_file, *out_file;
	unsigned char key[LONG_KEY_SIZE];
	struct stat in_stat, out_stat;
	int ret;

	set_prog_option(argc, argv, &prog_option);

	key_file = fopen(prog_option.key_filename, "rb");
	if (key_file == NULL)
		error("can\'t open key file `%s\'", prog_option.key_filename);
	if (fread(key, 1, LONG_KEY_SIZE, key_file) != LONG_KEY_SIZE) {
		fclose(key_file);
		error("error reading from key file `%s\'", prog_option.key_filename);
	}
	fclose(key_file);

	if ((prog_option.in_filename != NULL)
	    && (prog_option.out_filename != NULL)) {
		if (stat(prog_option.in_filename, &in_stat) == -1)
			error("can\'t open input file `%s'", prog_option.in_filename);
		if (S_ISDIR(in_stat.st_mode))
			error("can\'t open input file `%s'", prog_option.in_filename);

		if (stat(prog_option.out_filename, &out_stat) == -1) {
			if (in_stat.st_ino == out_stat.st_ino)
				error("`%s' and `%s' are the same file",
			    	  prog_option.in_filename, prog_option.out_filename);
		}
	}

	if (prog_option.in_filename == NULL)
		in_file = stdin;
	else
		in_file = fopen(prog_option.in_filename, "rb");
	if (in_file == NULL)
		error("can\'t open input file `%s\'", prog_option.in_filename);

	if (prog_option.out_filename == NULL)
		out_file = stdout;
	else
		out_file = fopen(prog_option.out_filename, "wb");
	if (out_file == NULL)
		error("can\'t open output file `%s\'", prog_option.out_filename);

	if (prog_option.verbose) {
		if (prog_option.crypt_dir == DES_ENCRYPT)
			printf("Encrypting `%s\' ", prog_option.in_filename);
		else
			printf("Decrypting `%s\' ", prog_option.in_filename);
		fflush(stdout);
	}
	
	ret = des_mode[prog_option.crypt_mode](out_file, in_file, key,
	                                       prog_option.crypt_dir);
	if (ret == 0) {
		if (prog_option.verbose)
			printf("done\n");
	} else {
		if (prog_option.verbose)
			printf("error\n");
		if (ret == 1)
			error("%s", strerror(errno));
		else
			error("file `%s\' is too short", prog_option.in_filename);
	}

	fclose(in_file);
	fclose(out_file);

	return 0;
}

void set_prog_option(int argc, char *argv[],
                     struct prog_option_t *prog_option)
{
	static struct option long_options[] = {
		{ "decrypt", no_argument, NULL, 'd' },
		{ "encrypt", no_argument, NULL, 'e' },
		{ "key", required_argument, NULL, 'k' },
		{ "mode", required_argument, NULL, 'm' },
		{ "verbose", no_argument, NULL, 'v' },
		{ "help", no_argument, NULL, 0 },
		{ "version", no_argument, NULL, 0 }
	};
	int option_index;
	char c;

	option_index = 0;
	while ((c = getopt_long(argc, argv, "dek:m:o:v",
	                        long_options, &option_index)) != -1) {
		switch (c) {
			case 'd':	/* decrypt */
				prog_option->crypt_dir = DES_DECRYPT;
				break;
			case 'e':	/* encrypt */
				prog_option->crypt_dir = DES_ENCRYPT;
				break;
			case 'k':	/* key */
				prog_option->key_filename = optarg;
				break;
			case 'm':	/* mode */
				if (strcmp(optarg, "ecb") == 0)
					prog_option->crypt_mode = MODE_ECB;
				else if (strcmp(optarg, "cbc") == 0)
					prog_option->crypt_mode = MODE_CBC;
				else if (strcmp(optarg, "ofb") == 0)
					prog_option->crypt_mode = MODE_OFB;
				else if (strcmp(optarg, "cfb") == 0)
					prog_option->crypt_mode = MODE_CFB;
				else
					error("bad mode: %s", optarg);
				break;
			case 'o':
				prog_option->out_filename = optarg;
				break;
			case 'v':
				prog_option->verbose = 1;
				break;
			case 0:
				switch (option_index) {
					case 5:	/* help */
						print_help(stdout);
						exit(0);
						break;
					case 6:	/* version */
						print_version(stdout);
						exit(0);
						break;
					default:	/* ??? */
						error("\
Unrecognized long option: %s\n\
This can\'t happen.", long_options[option_index].name);
						break;
				}
				break;
			case '?':
				fprintf(stderr, "\
Try `descrypt --help\' for more information.\n");
				exit(1);
			default:
				error("this can\'t happen.");
				break;
		}
	}
	if (prog_option->key_filename == NULL)
		error("you must specify a key file with `-k\' or `--key\'");
	if (prog_option->out_filename == NULL)
		prog_option->verbose = 0;
	if (optind < argc) {
		if (strcmp(argv[optind], "-") != 0)
			prog_option->in_filename = argv[optind];
		optind++;
		if (optind < argc)
			error("you may specify only one input file");
	}

	return;
}

void print_help(FILE *s)
{
	fprintf(s, "\
Usage: descrypt [OPTION]... -k KEY [FILE]\n\
descrypt encrypts or decrypts FILE and prints the resulting ciphertext\n\
or plaintext to a file or to standard output. Encryption is the default.\n\
\n\
  -d, --decrypt     specify decryption\n\
  -e, --encrypt     specify encryption (default)\n\
  -k, --key=KEY     use the first 8 bytes of file KEY as the key. This\n\
                      option may not be omitted.\n\
  -m, --mode=MODE   use the specified mode. MODE may be one of `ecb',\n\
                      `cbc' (default), `ofb', or `cfb'.\n\
  -o FILE           write output to FILE. Without this option, write to\n\
                      standard output.\n\
  -v, --verbose     display what is being done on standard output. This\n\
                      option is ignored if the -o option is omitted.\n\
      --help        display this usage message and exit\n\
      --version     display version information and exit\n\
\n\
If no FILE is given or if FILE is `-', read standard input.\n\
\n\
Report bugs to <david@bamsoftware.com>.\n\
");

	return;
}

void print_version(FILE *s)
{
	fprintf(s, "\
descrypt 0.1\n");

	return;
}

void error(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "error: ");
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);
	exit(1);
}

void warning(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "warning: ");
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);

	return;
}

